from .tools import dumps


VERTEX, EDGE = range(2)


class Properties(dict):
    """Properties are key/value pairs are are the association of a string
    with a value which can be anything among `int`, `float`, `str`, `list`
    and `dict`. Nesting `list` and `dict` is also accepted.

    This class keeps track of which values were set during the transaction to
    be able to update indexes. The original value of associated with a key can
    be retrieved with ``element.props.vanilla(key)``.

    .. note:: This class is not meant to be instaciated directly."""

    def __init__(self, vanilla, element):
        # store original values
        self._vanilla = vanilla
        self._element = element

    def items(self):
        return map(
            lambda key: (key, self.get(key, self._vanilla[key])),
            self.keys()
        )

    def keys(self):
        keys = list(super().keys())
        keys.extend(self._vanilla.keys())
        return keys

    def __setitem__(self, key, value):
        """Set property `key` to `value`"""
        self._element._dirty()
        super().__setitem__(key, value)

    def __getitem__(self, key):
        """Get the property associated with `key` of this element"""
        try:
            return super().__getitem__(key)
        except KeyError:
            return self._vanilla[key]

    def __getattr__(self, key):
        """Make it possible to use ``magic.property`` instead of
        ``magic['property']``"""
        try:
            return self[key]
        except KeyError:
            raise KeyError('no `{}` property key'.format(key))

    def vanilla(self, key):
        """Returns the original value associated with `key`

        "original value" means the value associated with `key` when the
        element was loaded. It raises `KeyError` if there is no value
        associated with that `key` **in the original version** of this
        properties."""
        return self._vanilla[key]

    def freeze(self):
        """Returns the new set of key/value pairs as a ``dict``"""
        new = dict(self._vanilla)
        new.update(self)
        return new


class Element:
    """Base class for Vertex and Edge.

    .. note:: This class is not meant to be instaciated directly.
    """

    def __init__(self, manager, identifier, label, properties):
        self._manager = manager
        self._label = label
        self._identifier = identifier
        # add a shortcut in da mix
        self.properties = self.props = Properties(properties, self)

    def identifier(self):
        """Returns the string identifier of the element"""
        return self._identifier

    def label(self):
        """Return the string label associated with thi label"""
        return self._label

    def _dirty(self):
        self._manager._txn._dirty.add(self)

    def delete(self):
        """Remove this useless element"""
        self._manager._txn._storage._delete(self)


class Vertex(Element):
    """Represent a vertex in graph database

    .. note:: This class is not meant to be instaciated directly."""

    def __init__(self,
                 manager,
                 identifier,
                 label,
                 outgoings,
                 incomings,
                 properties):
        super().__init__(manager, identifier, label, properties)
        self._outgoings = outgoings
        self._incomings = incomings

    def _iter_over_edges(self, kind, label=None):
        """Iterate over edges of `kind` possibly filtering them by `label`

        `kind` can be `"ougoings"` or `"incomings"`.

        :internal"""
        kind = '_%s' % kind
        kind = getattr(self, kind, dict())
        if label:
            try:
                identifiers = kind[label]
            except KeyError:
                raise StopIteration
            else:
                for identifier in identifiers:
                    yield self._manager._txn.edge.get(identifier)
        else:
            for identifiers in kind.values():
                for identifier in identifiers:
                    yield self._manager._txn.edge.get(identifier)

    def incomings(self, label=None):
        """Iterate over incoming edges possibly filtering by label

        label must be a string"""
        return self._iter_over_edges('incomings', label)

    def outgoings(self, label=None):
        """Iterate over outgoing edges possibly filtering by label

        label must be a string"""
        return self._iter_over_edges('outgoings', label)

    def freeze(self):
        return dict(
            identifer=self._identifier,
            label=self._label,
            properties=self.properties.freeze()
        )


class Edge(Element):
    """Represent an edge in graph database

    .. note:: This class is not meant to be instaciated directly."""

    def __init__(self, manager, identifier, start, label, end, properties):
        """Constructor is only used directly by a transaction to wrap an existing
        edge. Otherwise transaction objects use `Transaction.edge.create` to create
        a new ``Edge`` object"""
        super().__init__(manager, identifier, label, properties)
        self._start = start
        self._label = label
        self._end = end
        self._previous_properties = dict()

    def start(self):
        """Returns the vertex where this edge starts

        .. note:: If you only want to access the identifier of the *start node*
                  you can use ``Edge._start``

        .. warning:: You can not modifiy this value. If you want to do so
                     you will need to delete this edge and create a new one.
        """
        return self._manager._txn.vertex.get(self._start)

    def end(self):
        """Returns the vertex where this edge ends

        .. note:: If you only want to access the identifier of the *end node*
                  you can use ``Edge._end``

        .. warning:: You can not modifiy this value. If you want to do so
                     you will need to delete this edge and create a new one.
        """
        return self._manager._txn.vertex.get(self._end)

    def freeze(self):
        return dict(
            identifer=self._identifier,
            start=self._start,
            label=self._label,
            end=self._end,
            properties=self.properties.freeze()
        )
