from uuid import uuid4

from .tools import LazySliceGenerator

from .exceptions import NoElementFound


class BaseManager:

    def __init__(self, txn):
        self._txn = txn
        self._cache = dict()

    def _random_identifier(self):
        while True:
            identifier = uuid4().hex
            try:
                self.get(identifier)
            except NoElementFound:
                break
            else:
                continue
        return identifier


class VertexManager(BaseManager):

    def all(self):
        """Yields all vertices found in the database in lexical order based on identifers.

        .. warning:: this can impact performance on big graphs
        """
        yield from self._txn._storage._vertices_slice(self._txn).all()

    def count(self):
        """Return the total count of vertex"""
        return self._txn._storage._vertices_slice(self._txn).count()

    def slice(self, start, end):
        """Yield vertice between ``start`` and ``end`` index based on lexical ordering
        of the identifiers.

        .. note:: The vertex at index ``end`` is not included.

        .. note:: If the database was changed, the content of a slice might change
        """
        return self._txn._storage._vertices_slice(self._txn).slice(start, end)

    def label(self, label):
        generator = self._txn._storage.lookup(
            self._txn,
            'vertices_labels',
            'SS',
            label
        )

        def wrapper(pair):
            _, identifier = pair
            return self.get(identifier)

        return LazySliceGenerator(wrapper, generator)

    def get(self, identifier):
        return self._txn._storage.vertex(identifier, self, self._txn)

    def create(self, label=None):
        return self._txn._storage.create_vertex(label, self._txn)


class EdgeManager(BaseManager):
    
    def all(self):
        """Yields all edges found in the database in lexical order based on identifers.

        .. warning:: this can impact performance on big graphs
        """
        yield from self._txn._storage._edges_slice(self._txn).all()

    def count(self):
        """Return the total count of vertex"""
        return self._txn._storage._edges_slice(self._txn).count()

    def slice(self, start, end):
        """Yield edges between ``start`` and ``end`` index based on lexical ordering
        of the identifiers.

        .. note:: The edge at index ``end`` is not included.

        .. note:: If the database was changed, the content of a slice might
                  change between transactions.
        """
        return self._txn._storage._vertices_slice(self._txn).slice(start, end)

    def label(self, label):
        generator = self._txn._storage.lookup(self._txn, 'edges_labels', 'SS', label)

        def wrapper(pair):
            _, identifier = pair
            return self.get(identifier)

        return LazySliceGenerator(wrapper, generator)

    def get(self, identifier):
        return self._txn._storage.edge(identifier, self, self._txn)

    def create(self, start, label, end):
        return self._txn._storage.create_edge(start, label, end, self._txn)
