from pathlib import Path
from itertools import chain

from bsddb3.db import DB
from bsddb3.db import DBEnv
from bsddb3.db import DB_HASH
from bsddb3.db import DB_BTREE
from bsddb3.db import DB_CREATE
from bsddb3.db import DB_INIT_LOG
from bsddb3.db import DB_INIT_TXN
from bsddb3.db import DB_INIT_MPOOL
from bsddb3.db import DB_RECOVER_FATAL

from .tools import dumps
from .tools import loads

from .elements import Edge
from .elements import Vertex

from .managers import EdgeManager
from .managers import VertexManager

from .packing import pack
from .packing import unpack
from .tools import LazySliceGenerator
from .exceptions import NoElementFound


class Transaction:

    def __init__(self, storage):
        self._txn = storage._env.txn_begin()
        self.vertex = VertexManager(self)
        self.edge = EdgeManager(self)
        self._storage = storage
        self._dirty = set()

    def commit(self):
        for element in self._dirty:
            self._storage._save(element)
        self._txn.commit()

    def rollback(self):
        self._txn.abort()


class Storage:

    def __init__(self, path, cache_size=(0, 512), max_cache_size=(1, 0)):
        self._path = Path(path).absolute()
        # create directory (and its parents) if it doesn't exists
        if not self._path.exists():
            self._path.mkdir(parents=True)

        # init bsddb3
        self._env = DBEnv()
        self._env.set_cache_max(*max_cache_size)
        self._env.set_cachesize(*cache_size)
        flags = (
            DB_CREATE |
            DB_INIT_LOG |
            DB_INIT_TXN |
            DB_INIT_MPOOL
        )
        self._env.open(
            str(self._path),
            flags,
            0
        )

        # create vertices and edges k/v stores
        def new_store(name, method):
            txn = self._txn()
            flags = DB_CREATE
            elements = DB(self._env)
            elements.open(
                name,
                None,
                method,
                flags,
                0,
                txn=txn._txn
            )
            txn.commit()
            return elements

        self._vertices = new_store('vertices', DB_HASH)
        self._edges = new_store('edges', DB_HASH)
        self.vertices_labels = new_store('vertices_labels', DB_BTREE)
        self.edges_labels = new_store('vertices_labels', DB_BTREE)

    def _txn(self):
        return Transaction(self)

    def close(self):
        self._vertices.close()
        self._edges.close()
        self._env.close()

    def _save(self, element):
        txn = element._manager._txn._txn
        if isinstance(element, Vertex):
            value = (
                element._label,
                element._outgoings,
                element._incomings,
                element.properties.freeze()
            )
            store = self._vertices
            index = pack('SS', element._label, element._identifier)
            self.vertices_labels.put(index.encode('utf-8'), '', txn)
        else:
            value = (
                element._start,
                element._label,
                element._end,
                element.properties.freeze()
            )
            store = self._edges
            index = pack('SS', element._label, element._identifier)
            self.edges_labels.put(index.encode('utf-8'), '', txn)
        key = element._identifier.encode('utf-8')
        value = dumps(value)
        store.put(key, value, txn)

    def lookup(self, txn, database, fmt, *lookup):
        """Look up for entries based on `lookup`.

        The function only support string and integers. This must take into
        account how the queried database is constructed (see below). Querying
        happens over the key and returns the idenfiers.

        Label indexes ``'vertices_labels'`` and ``'edges_labels'`` expects
        only a label name so ``lookup`` equals to ``('some_label',)``.

        Properties indexes ``'vertices_properties'`` and ``'edges_properties'``
        expects an ``attribute_name`` and an ``attribute_value`` which can be
        a ``str`` or an ``int``.

        This can also be used to check for the existence of a entry, for
        instance, it's a quick way given the tuple (name, value, identifier)
        to know if the element associated with ``identifier`` has property
        ``name`` associated with ``value``. It's **probably** quickest than
        lloading the object.

        .. warning:: Don't do funky things with ``*key`` only what is
                     documented is supported (and it's for internal use only)!
        """
        database = getattr(self, database)
        cursor = database.cursor(txn._txn)
        index = pack(fmt[:len(lookup)], *lookup)
        next = cursor.set_range(index.encode('utf-8'))

        while next:
            values, _ = next
            values = unpack(fmt, values)
            # check that it's a good candidate
            if all(map(lambda x: x[0] == x[1], zip(lookup, values))):
                yield values
                next = cursor.next()
            else:
                break

    def create_vertex(self, label, txn):
        """Called by a vertex transaction manager to build a vertex.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.vertex._random_identifier()
        vertex = Vertex(txn.vertex, identifier, label, dict(), dict(), dict())
        txn.vertex._cache[identifier] = vertex
        vertex._dirty()
        return vertex

    def create_edge(self, start, label, end, txn):
        """Called by a edge transaction manager to build an edge.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.edge._random_identifier()
        edge = Edge(
            txn.edge,
            identifier,
            start._identifier,
            label,
            end._identifier,
            dict()
        )
        txn.edge._cache[identifier] = edge
        edge._dirty()
        # update start and end
        try:
            start._outgoings[edge._label].append(edge.identifier())
        except KeyError:
            start._outgoings[edge._label] = [edge.identifier()]
        try:
            end._incomings[edge._label].append(edge.identifier())
        except KeyError:
            end._incomings[edge._label] = [edge.identifier()]
        # mark vertices as dirty to be updated
        start._dirty()
        end._dirty()
        return edge

    def _delete(self, element):
        key = element._identifier.encode('utf-8')
        txn = element._manager._txn._txn
        if isinstance(element, Vertex):
            for edge in chain(element.outgoings(), element.incomings()):
                edge.delete()
            # delete this element from the store
            self._vertices.delete(key, txn)
        else:
            start = element.start()
            start._outgoings[element._label].remove(element._identifier)
            start._dirty()
            end = element.end()
            end._incomings[element._label].remove(element._identifier)
            end._dirty()
            # delete this element from the store
            self._edges.delete(key, txn)

    def _element(self, storage, ElementClass, identifier, manager, txn=None):
        try:
            element = manager._cache[identifier]
        except KeyError:
            key = identifier.encode('utf-8')
            element = storage.get(key, txn=manager._txn._txn)
            if not element:
                msg = 'Element with id=%s not found' % identifier
                raise NoElementFound(msg)
            element = loads(element)
            element = ElementClass(manager, identifier, *element)
            manager._cache[identifier] = element
        return element

    vertex = lambda s, *a, **kw: s._element(s._vertices, Vertex, *a, **kw)
    edge = lambda s, *a, **kw: s._element(s._edges, Edge, *a, **kw)

    def _slice(self, storage, ElementClass, txn):
        def generator():
            cursor = storage.cursor(txn=txn._txn)
            next = cursor.first()
            while next:
                yield next
                next = cursor.next()

        def wrapper(pair):
            identifier, element = pair
            identifier = identifier.decode('utf-8')
            element = loads(element)
            if ElementClass is Vertex:
                return ElementClass(txn.vertex, identifier, *element)
            else:
                return ElementClass(txn.edge, identifier, *element)

        return LazySliceGenerator(wrapper, generator())
    _vertices_slice = lambda s, *a, **kw: s._slice(s._vertices, Vertex, *a, **kw)
    _edges_slice = lambda s, *a, **kw: s._slice(s._edges, Edge, *a, **kw)
